/**
 *  Copyright (c) 2015, Facebook, Inc.
 *  All rights reserved.
 *
 *  This source code is licensed under the BSD-style license found in the
 *  LICENSE file in the root directory of this source tree. An additional grant
 *  of patent rights can be found in the PATENTS file in the same directory.
 */

'use strict';

var _inherits = require('babel-runtime/helpers/inherits')['default'];

var _classCallCheck = require('babel-runtime/helpers/class-call-check')['default'];

var _interopRequireDefault = require('babel-runtime/helpers/interop-require-default')['default'];

exports.__esModule = true;

var _react = require('react');

var _react2 = _interopRequireDefault(_react);

var _ExecuteButton = require('./ExecuteButton');

var _QueryEditor = require('./QueryEditor');

var _VariableEditor = require('./VariableEditor');

var _ResultViewer = require('./ResultViewer');

// import { DocExplorer } from './DocExplorer';

var _graphqlUtilities = require('graphql/utilities');

var _graphqlJsutilsFind = require('graphql/jsutils/find');

var _graphqlJsutilsFind2 = _interopRequireDefault(_graphqlJsutilsFind);

var _utilityFillLeafs = require('../utility/fillLeafs');

/**
 * GraphiQL
 *
 * This React component is responsible for rendering the GraphiQL editor.
 *
 * Props:
 *
 *   - fetcher: a function which accepts GraphQL-HTTP parameters and returns
 *     a Promise which resolves to the GraphQL parsed JSON response.
 *
 *   - schema: an optional GraphQLSchema instance. If one is not provided,
 *     GraphiQL will fetch one using introspection.
 *
 *   - defaultQuery: an optional GraphQL string to use instead of a
 *     blank screen when a query was not found in the local cache.
 *
 *   - getDefaultFieldNames: an optional function used to provide default fields
 *     to non-leaf fields which invalidly lack a selection set.
 *     Accepts a GraphQLType instance and returns an array of field names.
 *     If not provided, a default behavior will be used.
 *
 * Children:
 *
 *   - <GraphiQL.Logo> Replace the GraphiQL logo with your own.
 *
 *   - <GraphiQL.Toolbar> Add a custom toolbar above GraphiQL.
 *
 *   - <GraphiQL.Footer> Add a custom footer below GraphiQL Results.
 *
 */

var GraphiQL = (function (_React$Component) {
  _inherits(GraphiQL, _React$Component);

  function GraphiQL(props) {
    _classCallCheck(this, GraphiQL);

    _React$Component.call(this);

    // Ensure props are correct
    if (typeof props.fetcher !== 'function') {
      throw new TypeError('GraphiQL requires a fetcher function.');
    }

    var storage = window.localStorage;

    // Initialize state
    this.state = {
      schema: props.schema,
      response: null,
      editorFlex: storage.getItem('editorFlex') || 1,
      variableEditorOpen: storage.getItem('variableEditorOpen') === 'true' || false,
      variableEditorHeight: storage.getItem('variableEditorHeight') || 200,
      typeToExplore: null
    };

    // This is kept outside of state because it is just a cached value.
    this.queryString = storage.getItem('query') || props.defaultQuery || defaultQuery;

    this.variablesString = storage.getItem('variables');

    // Ensure only the last executed editor query is rendered.
    this.editorQueryID = 0;
  }

  // Configure the UI by providing this Component as a child of GraphiQL.

  /**
   * Inspect the query, automatically filling in selection sets for non-leaf
   * fields which do not yet have them.
   */

  GraphiQL.prototype.autoCompleteLeafs = function autoCompleteLeafs() {
    var _fillLeafs = _utilityFillLeafs.fillLeafs(this.state.schema, this.queryString, this.props.getDefaultFieldNames);

    var insertions = _fillLeafs.insertions;
    var result = _fillLeafs.result;

    if (insertions) {
      var editor = this.refs.queryEditor.getCodeMirror();
      editor.setValue(result);
      var added = 0;
      var markers = insertions.map(function (_ref) {
        var index = _ref.index;
        var string = _ref.string;
        return editor.markText(editor.posFromIndex(index + added), editor.posFromIndex(index + (added += string.length)), {
          className: 'autoInsertedLeaf',
          clearOnEnter: true,
          title: 'Automatically added leaf fields'
        });
      });
      setTimeout(function () {
        return markers.forEach(function (marker) {
          return marker.clear();
        });
      }, 7000);
    }
  };

  GraphiQL.prototype._fetchQuery = function _fetchQuery(query, variables, cb) {
    var _this = this;

    this.props.fetcher({ query: query, variables: variables }).then(cb)['catch'](function (error) {
      _this.setState({ response: JSON.stringify(error, null, 2) });
    });
  };

  GraphiQL.prototype._runEditorQuery = function _runEditorQuery() {
    var _this2 = this;

    this.editorQueryID++;
    var queryID = this.editorQueryID;

    this.autoCompleteLeafs();

    this._fetchQuery(this.queryString, this.variablesString, function (result) {
      if (queryID === _this2.editorQueryID) {
        _this2.setState({ response: JSON.stringify(result, null, 2) });
      }
    });
  };

  GraphiQL.prototype.componentDidMount = function componentDidMount() {
    var _this3 = this;

    if (!this.state.schema) {
      this._fetchQuery(_graphqlUtilities.introspectionQuery, null, function (result) {
        if (!result.data) {
          _this3.setState({ response: JSON.stringify(result, null, 2) });
        } else {
          _this3.setState({ schema: _graphqlUtilities.buildClientSchema(result.data) });
        }
      });
    }
  };

  GraphiQL.prototype._onHintInformationRender = function _onHintInformationRender(elem) {
    var onClickHintInformation = this._onClickHintInformation.bind(this);
    elem.addEventListener('click', onClickHintInformation);

    var onRemoveFn;
    elem.addEventListener('DOMNodeRemoved', onRemoveFn = function () {
      elem.removeEventListener('DOMNodeRemoved', onRemoveFn);
      elem.removeEventListener('click', onClickHintInformation);
    });
  };

  GraphiQL.prototype._onClickHintInformation = function _onClickHintInformation(event) {
    if (event.target.className === 'infoType') {
      var typeName = event.target.innerHTML;
      this.setState({
        typeToExplore: typeName
      });
    }
  };

  GraphiQL.prototype._onUpdate = function _onUpdate() {
    this.forceUpdate();
  };

  GraphiQL.prototype._onEdit = function _onEdit(value) {
    this.queryString = value;
    window.localStorage.setItem('query', value);
  };

  GraphiQL.prototype._onVariablesEdit = function _onVariablesEdit(value) {
    this.variablesString = value;
    window.localStorage.setItem('variables', value);
  };

  GraphiQL.prototype.render = function render() {
    var children = [];
    _react2['default'].Children.forEach(this.props.children, function (child) {
      children.push(child);
    });

    var logo = _graphqlJsutilsFind2['default'](children, function (child) {
      return child.type === GraphiQL.Logo;
    }) || _react2['default'].createElement(GraphiQL.Logo, null);

    var toolbar = _graphqlJsutilsFind2['default'](children, function (child) {
      return child.type === GraphiQL.Toolbar;
    });
    var footer = _graphqlJsutilsFind2['default'](children, function (child) {
      return child.type === GraphiQL.Footer;
    });

    var variableOpen = this.state.variableEditorOpen;
    var variableHeight = variableOpen ? this.state.variableEditorHeight : null;

    return _react2['default'].createElement(
      'div',
      { id: 'graphiql-container' },
      _react2['default'].createElement(
        'div',
        { className: 'topBar' },
        logo,
        _react2['default'].createElement(_ExecuteButton.ExecuteButton, { onClick: this._runEditorQuery.bind(this) }),
        toolbar
      ),
      _react2['default'].createElement(
        'div',
        {
          ref: 'editorBar',
          className: 'editorBar',
          onMouseDown: this.onResizeStart.bind(this)
        },
        _react2['default'].createElement(
          'div',
          { className: 'queryWrap', style: { flex: this.state.editorFlex } },
          _react2['default'].createElement(_QueryEditor.QueryEditor, {
            ref: 'queryEditor',
            schema: this.state.schema,
            value: this.queryString,
            onEdit: this._onEdit.bind(this),
            onHintInformationRender: this._onHintInformationRender.bind(this)
          }),
          _react2['default'].createElement(
            'div',
            { className: 'variable-editor', style: { height: variableHeight } },
            _react2['default'].createElement(
              'div',
              {
                className: 'variable-editor-title',
                style: { cursor: variableOpen ? 'row-resize' : 'n-resize' },
                onMouseDown: this.onVariableResizeStart.bind(this)
              },
              'Query Variables'
            ),
            _react2['default'].createElement(_VariableEditor.VariableEditor, {
              value: this.variablesString,
              onEdit: this._onVariablesEdit.bind(this)
            })
          )
        ),
        _react2['default'].createElement(
          'div',
          { className: 'resultWrap' },
          _react2['default'].createElement(_ResultViewer.ResultViewer, { ref: 'result', value: this.state.response }),
          footer
        )
      ),
      _react2['default'].createElement('div', { className: 'docExplorerWrap' })
    );
  };

  GraphiQL.prototype.onResizeStart = function onResizeStart(downEvent) {
    var _this4 = this;

    if (this.mouseMoveListener || !this.didClickDragBar(downEvent)) {
      return;
    }
    this.mouseOffset = downEvent.clientX - getLeft(downEvent.target);
    downEvent.preventDefault();
    this.mouseMoveListener = function (moveEvent) {
      if (moveEvent.buttons === 0) {
        _this4.onResizeEnd();
      } else {
        _this4.onResize(moveEvent);
      }
    };
    this.mouseUpListener = function () {
      _this4.onResizeEnd();
    };
    document.addEventListener('mousemove', this.mouseMoveListener);
    document.addEventListener('mouseup', this.mouseUpListener);
  };

  GraphiQL.prototype.onResizeEnd = function onResizeEnd() {
    window.localStorage.setItem('editorFlex', this.state.editorFlex);
    document.removeEventListener('mousemove', this.mouseMoveListener);
    document.removeEventListener('mouseup', this.mouseUpListener);
    this.mouseMoveListener = null;
    this.mouseUpListener = null;
  };

  GraphiQL.prototype.onResize = function onResize(event) {
    var editorBar = _react2['default'].findDOMNode(this.refs.editorBar);
    var leftSize = event.clientX - getLeft(editorBar) - this.mouseOffset;
    var rightSize = editorBar.clientWidth - leftSize;
    this.setState({ editorFlex: leftSize / rightSize });
  };

  GraphiQL.prototype.didClickDragBar = function didClickDragBar(event) {
    // Only for primary unmodified clicks
    if (event.button !== 0 || event.ctrlKey) {
      return false;
    }
    var target = event.target;
    // We use codemirror's gutter as the drag bar.
    if (target.className.indexOf('CodeMirror-gutter') !== 0) {
      return false;
    }
    // Specifically the result window's drag bar.
    var resultWindow = _react2['default'].findDOMNode(this.refs.result);
    while (target) {
      if (target === resultWindow) {
        return true;
      }
      target = target.parentNode;
    }
    return false;
  };

  GraphiQL.prototype.onVariableResizeStart = function onVariableResizeStart(downEvent) {
    var _this5 = this;

    var wasOpen = this.state.variableEditorOpen;
    var hadHeight = this.state.variableEditorHeight;
    if (!wasOpen) {
      this.setState({
        variableEditorOpen: true,
        variableEditorHeight: 30
      });
    }
    this.mouseOffset = downEvent.clientY - getTop(downEvent.target);
    downEvent.preventDefault();
    var didMove = false;
    this.mouseMoveListener = function (moveEvent) {
      didMove = true;
      if (moveEvent.buttons === 0) {
        _this5.onVariableResizeEnd();
      } else {
        _this5.onVariableResize(moveEvent);
      }
    };
    this.mouseUpListener = function (upEvent) {
      upEvent.preventDefault();
      if (!didMove) {
        var isOpen = !wasOpen;
        window.localStorage.setItem('variableEditorOpen', isOpen);
        _this5.setState({
          variableEditorOpen: isOpen,
          variableEditorHeight: hadHeight
        });

        // Simulate a window resize event so any CodeMirror instances become
        // properly rendered.
        window.dispatchEvent(new Event('resize'));
      } else {
        _this5.onVariableResizeEnd();
      }
    };
    document.addEventListener('mousemove', this.mouseMoveListener);
    document.addEventListener('mouseup', this.mouseUpListener);
  };

  GraphiQL.prototype.onVariableResizeEnd = function onVariableResizeEnd() {
    if (this.state.variableEditorHeight === 30) {
      window.localStorage.setItem('variableEditorHeight', 200);
      this.setState({ variableEditorOpen: false, variableEditorHeight: 200 });
    } else {
      window.localStorage.setItem('variableEditorHeight', this.state.variableEditorHeight);
    }
    document.removeEventListener('mousemove', this.mouseMoveListener);
    document.removeEventListener('mouseup', this.mouseUpListener);
    this.mouseMoveListener = null;
    this.mouseUpListener = null;
  };

  GraphiQL.prototype.onVariableResize = function onVariableResize(event) {
    var editorBar = _react2['default'].findDOMNode(this.refs.editorBar);
    var topSize = event.clientY - getTop(editorBar) - this.mouseOffset;
    var bottomSize = editorBar.clientHeight - topSize;
    if (bottomSize < 60) {
      bottomSize = 30;
    }
    this.setState({ variableEditorHeight: bottomSize });

    // Simulate a window resize event so any CodeMirror instances become
    // properly rendered.
    window.dispatchEvent(new Event('resize'));
  };

  return GraphiQL;
})(_react2['default'].Component);

exports.GraphiQL = GraphiQL;
GraphiQL.Logo = (function (_React$Component2) {
  _inherits(GraphiQLLogo, _React$Component2);

  function GraphiQLLogo() {
    _classCallCheck(this, GraphiQLLogo);

    _React$Component2.apply(this, arguments);
  }

  GraphiQLLogo.prototype.render = function render() {
    return _react2['default'].createElement(
      'div',
      { className: 'title' },
      this.props.children || _react2['default'].createElement(
        'span',
        null,
        'Graph',
        _react2['default'].createElement(
          'em',
          null,
          'i'
        ),
        'QL'
      )
    );
  };

  return GraphiQLLogo;
})(_react2['default'].Component);

// Configure the UI by providing this Component as a child of GraphiQL.
GraphiQL.Toolbar = (function (_React$Component3) {
  _inherits(GraphiQLToolbar, _React$Component3);

  function GraphiQLToolbar() {
    _classCallCheck(this, GraphiQLToolbar);

    _React$Component3.apply(this, arguments);
  }

  GraphiQLToolbar.prototype.render = function render() {
    return _react2['default'].createElement(
      'div',
      { className: 'toolbar' },
      this.props.children
    );
  };

  return GraphiQLToolbar;
})(_react2['default'].Component);

// Configure the UI by providing this Component as a child of GraphiQL.
GraphiQL.Footer = (function (_React$Component4) {
  _inherits(GraphiQLFooter, _React$Component4);

  function GraphiQLFooter() {
    _classCallCheck(this, GraphiQLFooter);

    _React$Component4.apply(this, arguments);
  }

  GraphiQLFooter.prototype.render = function render() {
    return _react2['default'].createElement(
      'div',
      { className: 'footer' },
      this.props.children
    );
  };

  return GraphiQLFooter;
})(_react2['default'].Component);

var defaultQuery = '# Welcome to GraphiQL\n#\n# GraphiQL is an in-browser IDE for writing, validating, and\n# testing GraphQL queries.\n#\n# Type queries into this side of the screen, and you will\n# see intelligent typeaheads aware of the current GraphQL type schema and\n# live syntax and validation errors highlighted within the text.\n#\n# To bring up the auto-complete at any point, just press Ctrl-Space.\n#\n# Press the run button above, or Cmd-Enter to execute the query, and the result\n# will appear in the pane to the right.\n\n';

function getLeft(initialElem) {
  var pt = 0;
  var elem = initialElem;
  while (elem.offsetParent) {
    pt += elem.offsetLeft;
    elem = elem.offsetParent;
  }
  return pt;
}

function getTop(initialElem) {
  var pt = 0;
  var elem = initialElem;
  while (elem.offsetParent) {
    pt += elem.offsetTop;
    elem = elem.offsetParent;
  }
  return pt;
}
// Temporarily disabled.
/* <DocExplorer
  ref="docExplorer"
  schema={this.state.schema}
  typeName={this.state.typeToExplore}
/> */